#!/usr/bin/perl
use JSON;
use CGI;
use utf8;
use Encode;

my %commands = (
	'cards'            => \&cards,
	'cat'              => \&cat,
	'copy'             => \&copy_func,
	'forward_vm'       => \&forward_vm,
	'get_cdr'          => \&get_cdr,
	'get_context_list' => \&get_context_list,
	'get_reports'      => \&get_reports,
	'ipchange'         => \&ipchange,
	'ls'               => \&ls,
	'ls_stat'          => \&ls_stat,
	'lwp'              => \&lwp,
	'mkdir'            => \&mkdir,
	'mv'               => \&mv,
	'noop'             => \&noop,
	'reload_ast'       => \&reload_ast,
	'reload_rpc'       => \&reload_rpc,
	'restart'          => \&restart,
	'restart_phone'    => \&restart_phone,
	'rm'               => \&rm,
	'save_context'     => \&save_context,
	'server_stats'     => \&server_stats,
	'set_localtime'    => \&set_localtime,
	'stat'             => \&stat,
	'sync'             => \&sync_config,
	'system_call'      => \&system_call,
	'system_call_md5'  => \&system_call_md5,
	'uae_routing'      => \&uae_routing,
	'upgres'           => \&upgres,
	'voicemail'        => \&voicemail,
	'undef'            => \&error

);

my $q    = new CGI;
my $json = JSON->new->utf8;

my $in = $json->allow_nonref->decode($q->param('input'));

#print STDERR Dumper($in);

# Make sure all incoming params are UTF8.
$in = utf8ify($in);

my $out = execute_cmd($in->{func}, $in->{params});

use Data::Dumper;

#print STDERR Dumper("out:",$out);

print $q->header;
print $json->allow_nonref->encode($out);

exit;

# UTF-8 encoding utility function
sub utf8ify {
	my $data = shift;
	my $encode = shift;
	if (ref($data) eq "HASH") {
		foreach my $key (keys %{$data}) {
			$data->{$key} = utf8ify($data->{$key}, $encode);
		}
		return $data;
	}
	elsif (ref($data) eq "ARRAY") {
		foreach my $item (@{$data}) {
			$item = utf8ify($item, $encode);
		}
		return $data;
	}
	else {
		my $is_utf8 = utf8::is_utf8($data);
		if ($encode) {
			return $data unless $is_utf8;
			$data = Encode::encode("UTF-8", $data);
		}
		else {
			return $data if $is_utf8;
			$data = Encode::decode("UTF-8", $data);

		}
		return $data;
	}
}

=head1 FUNCTIONS

=head2 exectute_cmd

Execute a function possibly w/ args
A lot the functions are system level:
cp, ls, stat, mv etc.

=cut

sub execute_cmd
{

	my $method = shift(@_);
	my @params = @{ shift(@_) };

	my $func  = $commands{$method};
	my $error = 0;
	unless ($func)
	{
		$func = $commands{"undef"};
		unshift(@params, $method);
		$error = 1;
	}

	my $data = $func->(@params);
	my $out;
	if ($error || $@)
	{
		$out = $data || $@;
	}
	else
	{
		$out->{value} = $data;
	}
	return $out;
}

=head2 error

=cut

sub error
{
	my $method = shift(@_);
	my @params = @{ shift(@_) };

	my $out;
	$out->{faultString}->{value} = "Unknown method: $method";
	$out->{is_fault} = 1;

	return $out;
}

=head2 cards

   Description:  Returns card information, alarms, etc
     Arguments:  NAME      IN/OUT  TYPE      DESCRIPTION
                 $card     in      string    Card number [optional]
       Returns:  Array of cards and info

=cut

sub cards
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: cards.xpl(@_)\n";
	close(LOG);
	my $card = shift;
	use strict;

	my %alarms = (
		none     => 0,
		recover  => 1,
		loopback => 2,
		yellow   => 4,
		red      => 8,
		blue     => 16,
		notopen  => 32,
	);

	my $ZT_SPANSTAT = -1066382838;

	my $zap = -e "/dev/dahdi" ? "dahdi" : "zap";
	open(CTL, "+</dev/$zap/ctl") || return $@ =
	  "Couldn't open /dev/$zap/ctl for rw: $!";

	my @cards;
	my @spans = $card ? ($card) : (1 .. 12);
	foreach my $span (@spans)
	{
		my $spaninfo = zt_spaninfo({ spanno => $span });
		ioctl(CTL, $ZT_SPANSTAT, $spaninfo) || last;
		$cards[ $span - 1 ] = { zt_spaninfo_unp($spaninfo) };

		$cards[ $span - 1 ]{alarms_text} =
		  [ grep { $cards[ $span - 1 ]{alarms} & $alarms{$_} } keys %alarms ];
	}
	close(CTL);

	sub zt_spaninfo
	{
		my %tmp = %{ $_[0] };

		return pack(
			"iA20A40iiiiiiiiiiii",
			$tmp{spanno},   $tmp{name},       $tmp{desc},
			$tmp{alarms},   $tmp{txlevel},    $tmp{rxlevel},
			$tmp{bpvcount}, $tmp{crc4count},  $tmp{ebitcount},
			$tmp{fascount}, $tmp{irqmisses},  $tmp{syncsrc},
			$tmp{numchans}, $tmp{totalchans}, $tmp{totalspans},
		);
	}

	sub zt_spaninfo_unp
	{
		my $tmp = shift;
		my @tmp = unpack("iA20A40iiiiiiiiiiii", $tmp);

		my $i = 0;
		return
		  map { $_ => $tmp[ $i++ ] }
		  qw/spanno name desc alarms txlevel rxlevel bpvcount crc4count
		  ebitcount fascount irqmisses syncsrc numchans totalchans totalspans/;
	}

	return \@cards;
}

=head2  ipchange

   Description:  Interfaces with the /var/adm/bin/ipchange.pl script to
                 change IP addresses.
     Arguments:  NAME      IN/OUT  TYPE      DESCRIPTION
                 $mode     in      string    STATIC | DHCP string
                 $settings in      ref       Configuration hash
       Returns:  None.

=cut

sub ipchange
{
	use strict;
	my ($mode, $settings) = @_;

	if (fork)
	{
		return 1;
	}

	else
	{
		open my $fh, "> /tmp/new_ip" or die $!;
		print $fh "mode=$mode\n";
		for (keys %$settings) { print $fh "$_ = $settings->{$_}\n" }
		close $fh;

		# the exec printing to STDOUT was corrupting the html/text data structure
		exec(
			"/var/www/rpc/surun /var/adm/bin/ipchange.pl /tmp/new_ip > /tmp/ipchange.log"
		);
	}
}

=head2  copy_func
   
   Description:  Downloads a remote file over HTTP to the specified path
   Arguments:    None.
   Returns:      1=success undef=faulure

=cut

sub copy_func
{
	log_args(@_);

	use strict;
	use LWP::Simple;
	use File::Basename;
	use Fcntl ':flock';
	use Digest::MD5;

	my ($url, $file, $md5, $astreload) = @{ shift(@_) };

	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: copy.xpl: $url, $file, $md5\n";
	close(LOG);

	unless ($url)
	{
		return $@ = "No URL specified";
	}
	unless ($file)
	{
		return $@ = "No file specified";
	}

	unless ($md5)
	{
		return $@ = "No md5 specified";
	}

	my $copyfile = $file . ".copy";

	# allowed path
	if ($file !~
		m;^(/data/\d*)?(/var/lib/asterisk|/etc/asterisk|/var/spool/asterisk|/tftpboot);
	  )
	{
		return $@ = "Cannot allow you to write to that directory";
	}

	# check if the md5 sum of the file matches the md5 of the file they want to sync
	if (open(FILE, "<$file"))
	{
		if (md5_hex(<FILE>) eq $md5)
		{
			close(FILE);
			return 1;
		}
		close(FILE);
	}

	# Make the directory if it does not exist
	my ($dirname) = dirname($file);
	my ($d);
	unless (-d $dirname)
	{    # Make the directory if needed
		my (@path_ary) = split(/\//, $dirname);
		foreach (1 .. scalar(@path_ary))
		{
			my ($p) = shift(@path_ary);
			$d .= "/$p";
			unless (-d $d)
			{
				mkdir($d);
			}
		}
	}

	# Create a lockfile to prevent simelaneous writes
	my $copylockfile = $copyfile . ".lck";

	open(COPYLOCKFILE, ">$copylockfile") || return $@ =
	  "COPY Failed: Can't open lockfile $copylockfile for writing: $!";
	flock(COPYLOCKFILE, LOCK_EX);

	# Execute the wget
	my ($rv) = getstore($url, $copyfile);
	if ($rv == 200)
	{

		# get the md5sum of the file
		if (!open(MD5FILE, $copyfile))
		{
			close(COPYLOCKFILE);
			unlink($copylockfile);
			return $@ = "COPY Failed: Can't open $copyfile: $!";
		}
		binmode(MD5FILE);

		my $localmd5 = Digest::MD5->new->addfile(*MD5FILE)->hexdigest;

		close(MD5FILE);

		if (($md5 !~ /\w{32}/) || ($md5 ne $localmd5))
		{

			# Remove the bad copy file
			if (-e $copyfile)
			{
				unlink($copyfile);
			}
			close(COPYLOCKFILE);
			unlink($copylockfile);
			return $@ =
			  "MD5 Failed: $rv=rv url=$url file=$copyfile remote_md5=$md5 local_md5=$localmd5";
		}

		close(COPYLOCKFILE);
		unlink($copylockfile);

		# The file arrived at the host machine and passed the md5 check
		# Now replace the real file with this new file
		# check for a lck file before preceeding

		my $lockfile = $file . ".lck";

		open(LOCKFILE, ">$lockfile") || return $@ =
		  "COPY Failed: Can't open lockfile $lockfile for writing: $!";
		flock(LOCKFILE, LOCK_EX);

		copy($copyfile, $file);

		if (-e $copyfile)
		{
			unlink($copyfile);
		}

		close(LOCKFILE);
		unlink($lockfile);

		if ($astreload)
		{
			reload_ast($file);
		}
		return (1);
	}
	else
	{
		close(COPYLOCKFILE);
		unlink($copylockfile);

		return $@ = "HTTP GET Failed: rv=$rv url=$url file=$file";
	}

}

=head2  cat

   Description:  Cats a file
   Arguments:    NAME      IN/OUT  TYPE      DESCRIPTION
                 $path     in      string    Path to cat
   Returns:      Data from file.

=cut

sub cat
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: cat.xpl(@_)\n";
	close(LOG);
	my $path = shift(@_);

	use strict;

	unless (open(FILE, "<$path"))
	{
		$@ = "Can't open $path for reading: $!";
		return undef;
	}

	return join("", <FILE>);
}

=head2  reload_ast

   Description:    Reloads asterisk
   Arguments:      None.
   Returns:        None.

=cut

sub reload_ast
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: reload_ast.xpl(@_)\n";
	close(LOG);
	use strict;

	my $file = shift;
	my $ret = undef;

	if (defined($file) && $file =~ m|^(/data/\d+)?/etc/asterisk/|)
	{
		if ($file =~ m|/sip\.conf$|)
		{
			$ret = `/var/www/rpc/astreload chan_sip.so`;
		}
		elsif ($file =~ m|/musiconhold\.conf$|)
		{
			$ret = `/var/www/rpc/astreload res_musiconhold.so`;
		}
		elsif ($file =~ m|/iax\.conf$|)
		{
			$ret = `/var/www/rpc/astreload chan_iax2.so`;
		}
		elsif ($file =~ m^/fonality/|/extensions\.conf$^)
		{
			$ret = `/var/www/rpc/astreload pbx_config.so`;
		}
		elsif ($file =~ m^/parking\.conf$|/features\.conf$^)
		{
			# there is no res_features in 1.6, just do standard reload
			$ret = `/var/www/rpc/astreload res_features.so`;
		}
		elsif ($file =~ m|/queues\.conf$|)
		{
			$ret = `/var/www/rpc/astreload app_queue.so`;
		}
		elsif ($file =~ m|/voicemail\.conf$|)
		{
			$ret = `/var/www/rpc/astreload app_voicemail.so`;
		}
		elsif ($file =~ m|/meetme\.conf$|)
		{
			$ret = `/var/www/rpc/astreload app_meetme.so`;
		}
		elsif ($file =~ m^/(?:alerts|agents|astwatch|pbxtra)\.conf$|\.xml$^)
		{
			# ignore
		}
		elsif ($file =~ /\.conf$/)
		{
			$ret = `/var/www/rpc/astreload`;
		}
	}
	else
	{
		# no specific file given, reload all
		$ret = `/var/www/rpc/astreload`;
	}

	return $ret;
}

=head2  cdr

   Description:    Return all CDR logs / delete CDR logs
   Arguments:      None.
   Returns:        None.

=cut

sub get_cdr
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: cdr.xpl(@_)\n";
	close(LOG);
	use strict;

	my $file = "/var/log/asterisk/cdr-csv/Master.csv";

	open(CSV, "<$file") || return "Can't open $file: $!";
	my $data = [ (<CSV>) ];
	close(CSV);

	return $data;
}

# does rpc-xml only allow one function per .code?
#sub rm_cdr
#{
#	use strict;
#	my($srv) = shift(@_);
#
#	my $file = "/var/log/asterisk/cdr-csv/Master.csv";
#
#	return unlink($file);
#}

=head2   noop

   Description:    Immediately returns 1
   Arguments:      None.
   Returns:        None.

=cut

sub noop
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: noop.xpl(@_)\n";
	close(LOG);
	return 1;
}

=head2 get_context_list

    Description: Get a list of all the contexts in the given file
    Args:        scalar_name_of_file_to_look_in
    Returns:     arrayref_of_context_names

=cut

sub get_context_list
{

	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: get_context_list.xpl(@_)\n";
	close(LOG);
	my $file = shift(@_) || '/etc/asterisk/extensions.conf';

	unless (-e $file)
	{
		if (-e "/etc/asterisk/$file")
		{
			$file = "/etc/asterisk/$file";
		}
		else
		{
			die "No such file $file: $!";
		}
	}

	my (@list, $contexts);

	foreach (@{ read_config($file) })
	{
		chomp($_);
		if (/^\[(.*)\]/)
		{
			$contexts->{$1} = 1;
			push(@list, $1);
		}
	}

	return (\@list);
}

=head2  mkdir

   Description:    Removes files and directories recursively
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                   $path     in      string    Path to remove
   Returns:        None.

=cut

sub mkdir
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: mkdir.xpl(@_)\n";
	close(LOG);
	my $path = shift(@_);

	use strict;

	if (-e $path)
	{
		return $@ = "$path already exists";
	}

	return mkdir($path);
}

=head2  ls_stat

   Description:    Returns stat() output for all files in a directory.
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                   $path     in      string    Path
   Returns:        Hashref of arrayrefs of info.

=cut

sub ls_stat
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: ls_stat.xpl(@_)\n";
	close(LOG);
	my $path = shift(@_);

	use strict;

	opendir(DIR, $path) || return $@ =
	  "ERROR: ls_stat: Can't opendir $path: $!";
	return {
		map { $_ => [ stat("$path/$_") ] }
		  grep { !-d "$path/$_" && !/^\./ } readdir(DIR)
	};
}

=head2  forward_vm

   Description:    Forwards a voicemail email to someone
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                   $req      in      arrayref  Array ref
                    [ email, file_path, from_name, from_email, subject, body, unbound_sid ]
   Returns:        None.

=cut

sub forward_vm
{
	log_args(@_);

	use strict;
	use Mail::Sender;

	my ($email, $file, $fromname, $fromemail, $subject, $body, $unbound_sid) =
	  @{ shift(@_) };

	# set default subject if necesarry
	$subject ||= "Voicemail from $fromname";

	# set default body if necesarry
	$body ||= <<"*END*";
$fromname has sent you a voicemail.

You can double click the attached file to hear it on your computer speakers.

*END*

	my $vmroot = "/var/spool/asterisk/voicemail/default/";

	if (-e "/etc/asterisk/unbound")
	{
		if ($unbound_sid !~ /^\d+$/)
		{
			return $@ =
			  "This is an UNBOUND server.  You are attempting to access this server using code that does not support Unbound servers";
		}

		$vmroot = "/var/spool/asterisk/voicemail/default-$unbound_sid/";
	}

	if ($email !~ /^[^\@\s\$]+\@(?:[a-z0-9-]+\.)+[a-z]+$/i)
	{
		return $@ = "Invalid email address";
	}

	$fromemail ||= "$fromname.nonexistant\@fonality.com";
	$fromemail =~ s/ /./g;

	$file = "$vmroot/$file";

	unless (-e $file)
	{
		return $@ = "No such file $file";
	}

	my $sender = new Mail::Sender {
		smtp => 'localhost',
		from => "\"$fromname\" <$fromemail>",
	};

	$sender->OpenMultipart(
		{
			to      => $email,
			subject => $subject,
		}
	);

	$sender->Body;
	$sender->SendEnc($body);

	$sender->Attach({ file => $file, });

	$sender->Close;

	return 1;

}

=head2  lwp

   Description:    does an lwp simple get on an ip and returns result
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                   $url      in      string    url to get
   Returns:        ip (string)

=cut

sub lwp
{
	use strict;
	use LWP::Simple;
	my $url = shift(@_);
	my $ip  = get($url);
	return ($ip);
}

=head2  ls

   Description:    Lists files or directories.
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                   $dir      in      string    Path
   Returns:        Array of files or directories.

=cut

sub ls
{
	log_args(@_);

	use strict;

	my @return;
	my ($path, $dir) = @{ shift(@_) };

	# if this is a file, just return it
	if (-e $path && !-d $path)
	{
		return [$path];
	}
	opendir(PATH, $path) || return $@ = "Can't read $path: $!";
	return [ grep { $dir ? -d "$path/$_" : !-d "$path/$_" } readdir(PATH) ];
}

=head2   mv

   Description:    Moves a file to a different path
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                   $from     in      string    Path to file
                   $to       in      string    Path to new file
   Returns:        None.

=cut

sub mv
{
	log_args(@_);
	use strict;

	my ($from, $to) = @{ shift(@_) };

	return rename($from, $to);
}

=head2  save_context

  Descriptions: Insert new or modified context into the configuration
          Args: scalar_name_of_context, arrayref_to_array_of_hashrefs
       Returns: 1 = success, 0 = failure

=cut

sub save_context
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: save_context.xpl(@_)\n";
	close(LOG);
	my ($context)         = shift(@_) || die "Context name missing";
	my ($new_context_ref) = shift(@_) || die "No context data given";
	my ($dir)                = '/etc/asterisk/fonality';
	my ($file)               = "${context}.conf";
	my ($include_index_file) = "${dir}/includes.conf";

	$context = strip_nasties($context);    # Clean out any bad characters

	unless (get_context($context))
	{    # If context is new, make an include for it
		open(INCL, ">>$include_index_file")
		  || die "Unable to open $include_index_file: $!";
		print INCL "\#include \"${dir}/${context}.conf\"\n";
		close(INCL);
	}

	unless (-d "${dir}/backups")
	{
		mkdir("${dir}/backups") || die "Can't mkdir: ${dir}/backups: $!";
	}

	copy("${dir}/${file}", "${dir}/backups/${file}.$$")
	  || die "Backup failed: $!";

	$file = "${dir}/${file}";

	open(FILE, ">$file") || die "Can't open: $file because: $!";
	print FILE "[$context]\n";
	foreach my $line (@{$new_context_ref})
	{
		my ($text) =
		  $line->{'extn'} . "," . $line->{'priority'} . "," . $line->{'app'};
		if (defined($line->{'comment'}))
		{
			$text .= $line->{'comment'};
		}
		$text .= "\n";
		print FILE "exten => $text";
	}
	close(FILE);
}

=head2  stat

   Description:    Returns stat() output for a file.
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                   $path     in      string    Path
   Returns:        Array of info.

=cut

sub stat
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: stat.xpl(@_)\n";
	close(LOG);
	my ($path) = @_;

	use strict;

	return [ stat($path) ];
}

=head2  restart_phone

   Description:    Restart a phone
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                   $extension     in      string    extension
   Returns:        Results 1 or undef

=cut

sub restart_phone
{
	use strict;
	use warnings;

	use LWP::UserAgent;
	use HTTP::Cookies;

	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: restart_phone.xpl(@_)\n";
	close(LOG);

	my ($ext_number) = @_;
	my $mac_add;

  READFILES: for ('internal', 'external')
	{
		open my $conf, '<', "/etc/asterisk/fonality/$_.conf" or die "$!";
		while (my $line = <$conf>)
		{
			my ($ext, $mac) = $line =~ m/(\d+),\d+,Dial\(SIP\/(\w+)/;
			if (($mac) && ($ext eq $ext_number))
			{
				$mac_add = $mac;

				close $conf;
				last READFILES;
			}
		}
		close $conf;
	}

	return undef unless $mac_add;

	# get ip by mac form asterisk
	my $out = `/usr/sbin/asterisk -rx "sip show peer $mac_add"`;
	my ($ip) = $out =~ m/Addr->IP\s+:\s(\S+)\s/;

	return undef unless $ip;

	# reboot astra phone by mac address
	if ($mac_add =~ /^00085D/)
	{
		my $browser    = LWP::UserAgent->new;
		my $cookie_jar = HTTP::Cookies->new();

		$browser->cookie_jar($cookie_jar);
		$browser->timeout(30);

		# default aastra username and password (and long realm from astra)
		$browser->credentials(
			"$ip:80",
			'Please enter User name and password',
			'admin' => '22222'
		);

		# set a cookies to cookie jar with first url
		my $resp = $browser->get("http://$ip/");
		$resp =
		  $browser->post("http://$ip/reset.html", [ 'resetOption' => '0' ]);

		if ($resp->content =~ m/Restarting/)
		{
			return 1;
		}
		else
		{
			return undef;
		}
	}
	else
	{
		return undef;
	}

}

=head2  grab_reports

   Description:    Grabs all call records
   Arguments:      None.
   Returns:        Hash with two arrays

=cut

sub get_reports
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: reporting.xpl(@_)\n";
	close(LOG);
	my ($req) = @_;
	my ($sid) = shift(@{$req});
	my ($uae) = shift(@{$req});

	use strict;
	use Date::Format;
	use File::Copy;

	open(LOGGER, ">>/tmp/log_pulling.log") or die "couldn't open a log file";

	print LOGGER time() . ": Starting the xpl portion of logging\n";

	my $reports;

	# This is now handled by the log pulling code
	#if ( -e "/etc/asterisk/unbound" && $sid ne 5221 )     # hack to prevent old logs code pulling from incorrect server
	#{
	#       return;
	#}

	# make a couple of dir's we need
	mkdir('/var/log/asterisk/real/');
	mkdir('/var/log/asterisk/nodata/');

	# list of file and locations to check for logs
	my $cdr_dir  = '/var/log/asterisk/cdr-csv/';
	my $cdr_file = 'Master.csv';

	my $queue_dir  = '/var/log/asterisk/';
	my $queue_file = 'queue_log';

	# backup directory
	my $backup_directory;

	# handle all cdr, or non queue calls
	my $queued_cdr_reports = check_for_queued_logs($cdr_dir);
	if ($queued_cdr_reports)
	{

		# We're here if there was a big log file that needed to be split up
		# we enter here for each file with *queued* in the name
		$reports->{'regular_reports'} =
		  grab_regular_reports($queued_cdr_reports);

		$queued_cdr_reports =~ m|([^/]+)$|;

		unless (rename($queued_cdr_reports, "/var/log/asterisk/real/$1"))
		{
			$reports->{'regular_reports'} = undef;
			print LOGGER
			  "Unable to rename file $queued_cdr_reports /var/log/asterisk/real/$1 :  $! $@\n";
		}
	}
	else
	{

		# no queued files, let's look at the origional file
		my $call_logs = get_new_logs($cdr_dir, $cdr_file);
		if ($call_logs)
		{
			$reports->{'regular_reports'} = grab_regular_reports($call_logs);

			$call_logs =~ m|([^/]+)$|;

			unless (rename($call_logs, "/var/log/asterisk/real/$1"))
			{
				$reports->{'regular_reports'} = undef;
				print LOGGER
				  "Unable to rename file $call_logs /var/log/asterisk/real/$1 :  $! $@\n";
			}
		}
	}

	#handle all queue calls
	my $queued_queue_reports = check_for_queued_logs($queue_dir);
	if ($queued_queue_reports)
	{
		$reports->{'queue_reports'} = grab_queue_reports($queued_queue_reports);

		$queued_queue_reports =~ m|([^/]+)$|;

		unless (rename($queued_queue_reports, "/var/log/asterisk/real/$1"))
		{
			$reports->{'queue_reports'} = undef;
			print LOGGER
			  "Unable to rename file $queued_queue_reports /var/log/asterisk/real/$1 :  $! $@\n";
		}
	}
	else
	{
		my $call_logs = get_new_logs($queue_dir, $queue_file);
		if ($call_logs)
		{
			$reports->{'queue_reports'} = grab_queue_reports($call_logs);

			$call_logs =~ m|([^/]+)$|;

			unless (rename($call_logs, "/var/log/asterisk/real/$1"))
			{
				$reports->{'queue_reports'} = undef;
				print LOGGER
				  "Unable to rename file $call_logs /var/log/asterisk/real/$1 :  $! $@\n";
			}
		}
	}

	print LOGGER time() . ": Stopping the xpl portion of logging\n";

	# if this is a Unified Agent Edition server,
	# grab call recordings for directory and add to the output variable as an array
	if ($uae)
	{
		print LOGGER time()
		  . ": Unified Agent Edition : Starting call file importing\n";

		my $path = "/var/spool/asterisk/monitor";
		opendir(DIR, $path) or die "Couldn't open $path: $!";

		my @calls = grep { !/^\.\.?$/ } readdir(DIR);

		$reports->{'calls'} = \@calls;

		print LOGGER time()
		  . ": Unified Agent Edition : Stopping call file importing\n";
		closedir(DIR);
	}

	close(LOGGER);
	return $reports;

	# just as it sounds, look in a given $dir for filed with queued in the file name
	sub check_for_queued_logs
	{
		my $dir = shift(@_);

		opendir(DIR, "$dir");
		my @files = readdir(DIR);
		closedir(DIR);

		foreach my $file (@files)
		{
			if ($file =~ /queued/)
			{
				return "$dir$file";
			}
			elsif ($file =~ /queue_log\.\d{1,}/)
			{
				return "$dir$file";
			}
		}
		return undef;
	}

	sub grab_regular_reports
	{
		my $file = shift(@_);

		if (-e "$file")
		{
			my $cleaned_up_reports;
			my $time = time;

			my $devices = devices_list();

			open(CDR, "<$file") || die "Can't open csv: $!";
			while (<CDR>)
			{

				# strip out nasty characters
				s/[^\x20-\x7e]//g;

				# strip out excess quotes
				s/""//g;

				# append a comma to our string for faster parsing
				$_ .= ",";

				# need to empty out our @fields array
				my @fields = ();

				# read each element into an array
				s/(?:"(.*?)",|([^,]*),)/push @fields, $+/eg;

				# remove the 11th element of the array
				splice(@fields, 10, 1);

				push(@fields, $fields[1]);

				# remove the 15th element of the array
				# leave blank for the origsrc and just get the jitter info
				splice(@fields, 15, 1);

				# replace the sip/<mac>-<random garbage>
				# with the extension number
				if (length($fields[1]) > 6)
				{
					my ($mac) = ($fields[5] =~ /^"?([^-]+)/);
					if ($devices->{$mac})
					{
						$fields[1] = "$devices->{$mac}";
					}
				}

				$cleaned_up_reports .=
				  sprintf(
					"\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",%d,%d,\"%s\",%d,\"\",\"%s\"\n",
					@fields);
			}
			close(CDR);

			return $cleaned_up_reports;
		}
	}

	sub grab_queue_reports
	{
		my $file = shift(@_);
		my @enter;
		my @abandon;
		my @callconnect;

		my $reports;

		if (-e "$file")
		{
			my @enter;
			my @abandon;
			my @callconnect;

			open(LOG, "$file");
			while (my $line = <LOG>)
			{
				chomp($line);
				if ($line =~
					/(\d+)\|(.+?)\|(.+?)\|NONE\|ENTERQUEUE\|(.*?)\|(.*)/)
				{
					push(@enter,
						    qq{","$2","$3","}
						  . ($4 ? "$4" : "NULL") . qq{","}
						  . ($5 ? "$5" : "NULL") . qq{","}
						  . time2str("%Y-%m-%d %H:%M:%S", $1)
						  . q{"});
				}
				elsif ($line =~
					/(\d+)\|(.+?)\|(.+?)\|NONE\|ABANDON\|(\d*?)\|(\d*?)\|(\d*)/)
				{
					push(@abandon,
						    qq{","$2","$3","}
						  . ($4 || 0) . qq{","}
						  . ($5 || 0) . qq{","}
						  . ($6 || 0) . qq{","}
						  . time2str("%Y-%m-%d %H:%M:%S", $1)
						  . qq{"});
				}
				elsif ($line =~ /(\d+)\|(.+?)\|(.+?)\|(.*?)\|AGENTDUMP\|/)
				{
					push(@callconnect,
						qq{","$2","$3","$4","agentdump","NULL","NULL","NULL","}
						  . time2str("%Y-%m-%d %H:%M:%S", $1)
						  . qq{"});
				}
				elsif ($line =~ /(\d+)\|(.+?)\|(.+?)\|(.*?)\|CONNECT\|(\d+)/)
				{
					push(@callconnect,
						    qq{","$2","$3","$4","connect","$5","NULL","NULL","}
						  . time2str("%Y-%m-%d %H:%M:%S", $1)
						  . q{"});
				}
				elsif ($line =~
					/(\d+)\|(.+?)\|(.+?)\|(.*?)\|COMPLETEAGENT\|(\d+)\|(\d+)[|]{0,1}(\d*)/
				  )
				{
					push(@callconnect,
						    qq{","$2","$3","$4","completeagent","$5","$6","}
						  . ($7 ? "$7" : "NULL") . q{","}
						  . time2str("%Y-%m-%d %H:%M:%S", $1)
						  . q{"});
				}
				elsif ($line =~
					/(\d+)\|(.+?)\|(.+?)\|(.*?)\|COMPLETECALLER\|(\d+)\|(\d+)[|]{0,1}(\d*)/
				  )
				{
					push(@callconnect,
						    qq{","$2","$3","$4","completecaller","$5","$6","}
						  . ($7 ? "$7" : "NULL") . qq{","}
						  . time2str("%Y-%m-%d %H:%M:%S", $1)
						  . qq{"});
				}
			}

			close(LOG);

			if (   scalar(@enter) == 0
				&& scalar(@abandon) == 0
				&& scalar(@callconnect) == 0)
			{
				$file =~ /(.*)\/.*\.(.*)$/;
				my $file_to_unlink =
				  $1 . "/" . $backup_directory . "/queue_log." . $2;
				rename($file_to_unlink,
					"/var/log/asterisk/nodata/queue_log.$2");
			}
			else
			{
				if (scalar(@enter))
				{
					$reports->{'enter'} = \@enter;
				}

				if (scalar(@abandon))
				{
					$reports->{'abandon'} = \@abandon;
				}

				if (scalar(@callconnect))
				{
					$reports->{'callconnect'} = \@callconnect;
				}
			}

			return $reports;
		}
	}

	sub devices_list
	{
		my %devices = ();

		my @files = (
			'/etc/asterisk/fonality/internal.conf',
			'/etc/asterisk/fonality/external.conf'
		);

		foreach my $file (@files)
		{
			open(FILE, "<$file");
			for my $line (<FILE>)
			{
				if ($line =~
					/\s*exten\s*=>\s*(.*),.*,Dial\(((?:SIP|Zap|IAX2)[^,]+),/)
				{
					$devices{$2} = $1;
				}
			}
			close(FILE);
		}

		return \%devices;
	}

	sub get_new_logs
	{
		my $path = shift(@_);
		my $file = shift(@_);

		my $origional_file = "$path$file";

		my $new_file = $origional_file . ".queued." . time();

		# used to place log files in a directory on a week by week basis
		my @time = localtime();
		$backup_directory = sprintf("%02d-%02d-%02d",
			substr($time[5] + 1900, 2),
			$time[4] + 1,
			int(($time[3] - 1) / 7) * 7 + 1);
		mkdir("$path$backup_directory");

		if (-e $origional_file)
		{

			# need to move the file out of the way and get asterisk going again
			rename($origional_file, $new_file);

			# lets restart the asterisk logger just to make sure we're good to go with
			# opening a new log file
			my $output = `/usr/sbin/asterisk -r -x 'logger reload'`;
			print LOGGER "Output: $output\n";

			# find out how big this file is
			my $size = (stat($new_file))[7];

			# if the file passed in is smaller then .8MB then this file is good enough
			if ($size <= 800000)
			{
				my $destination_file = $origional_file . ".queued." . time();

				# The resultant file is returned to be processed, it's renamed with .queued. in case anything goes wrong
				rename($new_file, $destination_file);

				# The resultant file is uses as a copy to remain on the pbxtra/trixbox

				copy($destination_file,
					"$path$backup_directory/$file." . time());
				return $destination_file;
			}
			else
			{

				# that pesky file is just too darn big, let's go ahead and
				# break it down into smaller pieces, and return just one piece

				my $file_ending = 1;
				my $size;

				open(FILE, "<$new_file");

				$origional_file =~ /(.*\/)(.*)/;
				my $tmpfile = $1 . "$2.csv.queued.";

				open(TEMPFILE, ">$tmpfile$file_ending")
				  or die print "Problems: $!\n";

				# as long as there's data i the file
				while (<FILE>)
				{
					my $line = $_;

					# clean it up
					chomp($line);
					print TEMPFILE "$line\n";

					# record the size
					$size += length($line);

					# file too big? let's open a new one!
					if ($size >= (800000))
					{
						close(TEMPFILE);
						$file_ending++;
						open(TEMPFILE, ">$tmpfile$file_ending");
						$size = 0;
					}
				}
				close(TEMPFILE);
				close(FILE);

				# this file remains on the pbxtra/trixbox for history
				my $destination_file = "$path$backup_directory/$file." . time();

				rename($new_file, $destination_file);

				print LOGGER "Returning: $tmpfile$file_ending\n";

				# we return the last queued file put together
				return "$tmpfile$file_ending";
			}
		}
	}
}

=head2  reload_rpc

   Description:    Creates new .xpl files and restarts apache
   Arguments:      None.
   Returns:        None.

=cut

sub reload_rpc
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: reload_rpc.xpl(@_)\n";
	close(LOG);
	use strict;
	return `/var/www/rpc/suidrpc`;
}

=head2  sync_config

   Description:    Syncs configuraiton files on the client machine
   Arguments:      None.
   Returns:        None.

=cut

sub sync_config
{
	log_args(@_);
	use strict;
	use File::Copy;
	use Fcntl ':flock';
	use Digest::MD5 qw/md5_hex/;

	my ($file, $md5, $data, $path, $noreload, $mosted_ok) = @{ shift(@_) };

	# Check for mosted support - we pass an additional dummy parameter to verify that the
	# call is from a modern RPC lib.
	if (-e "/etc/asterisk/unbound" && !$mosted_ok)
	{
		return $@ =
		  "This is an UNBOUND server.  You are attempting to access this server using code that does not support Unbound servers";
	}

	my $origfile = $file;
	unless ($file)
	{
		return $@ = "No file specified";
	}
	unless ($md5)
	{
		return $@ = "No md5 specified";
	}
	$file =~ s|^/etc/asterisk/||;

	# allowed path(s)
	if (   $path
		&& $path !~
		m;^(/data/\d*/*)?(/usr/local/fonality|/usr/local/ircd/etc|/tftpboot|/etc/asterisk|/var/lib/asterisk|/etc/postfix|/usr/local/tftpd(?:/|$)|/var/spool/asterisk/outgoing(?:/|$));
	  )
	{
		return $@ = "Cannot allow you to write to that directory";
	}

	if (!$path && $file =~ m;^(/data/\d+/)(.*);)
	{
		$path = $1;
		$file = $2;
	}

	$path ||= "/etc/asterisk";

	if ($file =~ /(?:tftp|smg_bri|wanpipe)/)
	{
		$path = "";
	}

	# has $data when we want to sync the actual file
	if (defined($data))
	{

		# check if the md5 sum of the file matches the md5 of the file they want to sync
		if (open(FILE, "<$path/$file"))
		{
			if (md5_hex(<FILE>) eq $md5)
			{
				close(FILE);
				return 1;
			}
			close(FILE);
		}

		# make directories for any directories that are not made
		my $temp;

		# go through each directory and create it if it hasn't already been
		foreach (
			(split(m|/|, "$path/$file"))[ 0 .. "$path/$file" =~ tr|/|| - 1 ])
		{
			$temp .= "$_/";
			if (-e $temp)
			{
				unless (-d $temp)
				{
					return $@ = "$temp is not a directory";
				}
			}
			else
			{
				mkdir($temp) || return $@ = "Can't mkdir($temp): $!";
			}
		}

		# if data is empty but file exists, don't sync
		if ((length($data) == 0) && (-e "$path/$file"))
		{
			return $@ = "Can't sync an empty file - file exists: $path/$file.";
		}

		# let's sync file if data is not empty or if file doesn't exist yet

		# use a semaphore in case the this process and another
		# do the open() before either perform a flock()
		my $semaphore = "$path/$file.lck";
		open(SEMAPHORE, ">$semaphore") || return $@ = "Can't open semaphore $semaphore for writing: $!";

		flock(SEMAPHORE, LOCK_EX);

		# backup the file we're going to overwrite
		copy("$path/$file", "$path/bak/$file." . time() . ".$$");

		if (!open(FILE, ">$path/$file"))
		{
			# remove lock before returning
			close(SEMAPHORE);
			unlink($semaphore);
			return $@ = "Unable to write to $path/$file: $!";
		}
		print FILE $data;
		close FILE;

		close(SEMAPHORE);
		unlink($semaphore);

		if (!$noreload && $origfile =~ m|^(/data/\d+)?/etc/asterisk/|)
		{
			reload_ast($origfile);
		}

		return 1;
	}

	else
	{
		# we get to this else because we want to check md5 sum only
		if (-e "$path/$file")
		{

			# check if the md5 sum of the file matches the md5 of the file they want to sync
			open(FILE, "<$path/$file") || return $@ =
			  "Can't read $path/$file: $!";
			if (md5_hex(<FILE>) eq $md5)
			{
				close(FILE);
				return 1;
			}
			close(FILE);
			return 0;
		}
		else
		{
			return 0;
		}
	}
}

=head2  system_call_md5

   Description:    Run and return system command
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                   $card     in      string    Card number [optional]
   Returns:        Results from system command

=cut

sub system_call_md5
{
	my $commandurl = shift(@_);
	my $comm       = @{$commandurl}->[0];
	my ($command, $url) = @{$comm};

	use strict;
	use Digest::MD5;

	my $ret = "";

	my $lwpget     = `/usr/bin/lwp-request $url`;
	my $md5command = Digest::MD5::md5_hex($command);
	if ($lwpget eq $md5command)
	{
		$ret = `$command`;
	}
	else
	{
		$ret = "FAIL: $md5command NE FROM $url: \n$lwpget";
	}

	return $ret;
}

=head2 set_localtime

  Description: locally copies the time zone file provided to /etc/localtime
               to update the servers localtime.
         Args: file path
      Returns: 1 on success or die "why" on error

=cut

sub set_localtime
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: set_localtime.xpl(@_)\n";
	close(LOG);
	my $file_aref =
	  shift;    # file is passed in an arrayref so deref for the first val

	my $file   = $file_aref->[0];
	my $mosted = $file_aref->[1];
	my $prefix;

	if ($mosted)
	{
		$prefix = "/data/$mosted";
	}

	use File::Copy;

	#
	#  Verify that the person is trying to copy a time zone file
	#
	if ($file !~ /^\/usr\/share\/zoneinfo/)
	{
		die
		  "Unable to copy local time file '$file' to '$prefix/etc/localtime'. Reason: "
		  . "file '$file' is not recognized as a time zone file.";
	}

	#
	#  Get the file data from the new timezone file
	#
	if (system("/var/adm/bin/set_localtime","$prefix/etc/localtime",$file) == 0)
	{
		return 1;
	}
	else
	{
		die
		  "Unable copy local time file '$file' to '$prefix/etc/localtime'. Reason: unable to read file '$file'";
	}
}

=head2  rm

   Description:    Removes files and directories recursively
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                   $path     in      string    Path to remove
   Returns:        None.

=cut

sub rm
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: rm.xpl(@_)\n";
	close(LOG);
	my ($path) = @_;

	use strict;


	unless (-e $path)
	{
		return 1;
	}

	foreach (recurse($path))
	{
		-d $_ ? rmdir($_) : unlink($_);
	}
	return -d $path ? rmdir($path) : unlink($path);
}

=head2  ur

   Description:    Update Responder Interface
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                   $command  in      string    UR command
		           $hash     in      string    hash key for server
   Returns:        status

=cut

sub upgres
{
	my ($req) = @_;
	my ($server_id, $command, $hash, $beta) = @{$req};

	use strict;

	if ($beta != 1)
	{
		$beta = 0;
	}

	my $ret =
	  `/usr/bin/lwp-request "http://update.fonality.com/upgres.cgi?server_id=s$server_id&hash=$hash&beta=$beta&action=$command"`;

	return $ret;
}

=head2  voicemail

   Description:    Send a hello message
   Arguments:      hash with an extension.
   Returns:        None.

=cut

sub voicemail
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: voicemail.xpl(@_)\n";
	close(LOG);
	use strict;
	my ($req) = @_;

	my $extension = $req->{extension};
	my $server_id = $req->{server_id};

	return $@ = "Extension ($extension) must be numeric"
	  if $extension !~ /^\d+$/;
	my $vmdir;
	if ($server_id)
	{
		$vmdir = "/var/spool/asterisk/voicemail/default-$server_id/$extension";
	}
	else
	{
		$vmdir = "/var/spool/asterisk/voicemail/default/$extension";
	}

	# create a hash of hashrefs, keys are folder names
	my %folders;

	if (-e $vmdir)
	{

		# get the folders for the extension
		opendir(EXTEN, $vmdir) || return $@ = "Couldn't open vmdir $vmdir: $!";
		foreach my $folder (grep { !/^\./ && -d "$vmdir/$_" } readdir(EXTEN))
		{
			$folders{$folder} = {};

			# get all the files in the dir
			opendir(FOLDER, "$vmdir/$folder") || return $@ =
			  "Couldn't open folder $vmdir/$folder: $!";
			foreach my $file (grep { /\.txt$/ } readdir(FOLDER))
			{
				$folders{$folder}{$file} = {};

				open(TXT, "<$vmdir/$folder/$file") || return $@ =
				  "Couldn't open text $vmdir/$folder/$file: $!";
				while (<TXT>)
				{
					next if /^\s*;|^\s*$/;

					$folders{$folder}{$file}{$1} = $2
					  if /^\s*(\S+)\s*=\s*(.*)$/;
				}
				close(TXT);
			}
			closedir(FOLDER);
		}
		closedir(EXTEN);
	}
	else
	{
		return $@ =
		  "vmdir $vmdir doesn't exist or we do not have correct permissions";
	}

	return \%folders;
}

=head2  system_call

   Description:    Run and return system command
   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
                  $card     in      string    Card number [optional]
   Returns:        Results from system command

=cut

sub system_call
{
	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() . ": $$: system_call.xpl(@_)\n";
	close(LOG);
	my ($command) = @_;

	use strict;

	my $ret = `$command`;

	return $ret;
}

=head2   restart

   Description:    Restarts asterisk
   Arguments:      array of restart command line options.
   Returns:        None.

=cut

sub restart
{
	log_args(@_);
	use strict;
	my ($arg_arrayref) = shift(@_);
	my ($restart_type) = shift(@{$arg_arrayref});

	system("/var/www/rpc/astrestart $restart_type");
	return 1 if ($? == 0);    # It worked okay
	return 0;
}

=head2   server_stats

   Description:    current server stats (cpu, mem).
   Arguments:      None.
   Returns:        Array of Server stats.

=cut

sub server_stats
{
	log_args(@_);

	# Felix Nilam - 12/21/2009
	# fix the input data
	#my ($howlong, $type) = @_;
	#my $data = [$howlong, $type];
	my ($data) = @_;
	my ($howlong, $type) = @{$data};

	my $time = time();

	my $start = $time - 1;
	if ($howlong eq "live")
	{
		$start -= 60;
	}
	elsif ($howlong eq "hour")
	{
		$start -= (1 * 60 * 60);
	}
	elsif ($howlong =~ /^(\d+)min$/)
	{
		$start -= ($1 * 60);
	}
	elsif ($howlong eq "sixhour")
	{
		$start -= (6 * 60 * 60);
	}
	elsif ($howlong eq "daily")
	{
		$start -= (24 * 60 * 60);
	}
	elsif ($howlong eq "weekly")
	{
		$start -= (7 * 24 * 60 * 60);
	}
	elsif ($howlong eq "monthly")
	{
		$start -= (30 * 24 * 60 * 60);
	}

	my $end = $time - 2;

	use strict;
	use lib qw( /usr/local/rrdtool-1.2.13/lib/perl );
	use lib qw( /usr/local/rrdtool-1.2.15/lib/perl );
	use RRDs;

	#
	# types:
	#   0 = server_stats
	#   1 = call_stats
	#   2 = network
	#
	my @DB = qw|
	  /var/lib/asterisk/server_stats.rrd
	  /var/lib/asterisk/call_stats.rrd
	  /var/lib/asterisk/network_stats.rrd
	  |;
	my $dst = "AVERAGE";

	my ($dbstart, $step, $names, $data) =
	  RRDs::fetch($DB[$type], "--start=$start", "--end=$end", $dst);

	if ($howlong eq "hour" || $howlong =~ /^(\d+)min$/)
	{
		for (my $i = $#$data + 1 ; $i >= 0 ; $i--)
		{

			# $1 || 60 (defined minutes or hour)
			# that many minutes over 15, for example, 60 / 15 = 4
			# 10 * that num == 40, so only show 1 point for every 40
			if ($i % int(10 * (($1 || 60) / 15)))
			{
				splice(@{$data}, $i, 1);
			}
		}
	}

	return [ $data, $start, $end ];
}

=head2   uae_routing

   Description:    Manage Call Routing Data
   Arguments:      operator (add, modify, get, delete, clear),
		    server ID,
		    array of phone numbers, extensions, and descriptions
   Returns:        Success/failure/data hash

=cut

sub uae_routing
{
	my ($req, $phones, $exts, $descs) = @_;
	my ($sid) = shift(@{$req});
	my ($opt) = shift(@{$req});

	use strict;
	use File::Copy;
	use BerkeleyDB;

	my $env = new BerkeleyDB::Env -Flags => DB_CREATE;

	my $db = new BerkeleyDB::Hash(
		-Filename => "/var/lib/asterisk/call_routing.db",
		-Flags    => DB_CREATE,
		-Env      => $env
	) or die "Could not open DB";

	open(LOGGER, ">>/tmp/test_routing.log")
	  or die "couldn't open a log file"
	  or die "Can't open file: $!";

	# adding
	if ($opt eq 'add' || $opt eq 'modify')
	{

		# loop through phones, put into db indexed by phone number
		# (assumes array of phones, exts, desc are of equal length)
		while (@{$phones})
		{

			my $index  = shift(@{$phones});
			my $record = shift(@{$exts}) . ":" . shift(@{$descs});

			my $status = $db->db_put($index, $record);

			print LOGGER localtime()
			  . ":: add status: $status, adding: $index : $record\n";

			# if error
			if ($status)
			{
				my $result->{'faultString'} = "Could not add ($index, $record)";

				$db->db_close();
				close(LOGGER);

				return $result;
			}
		}
	}
	elsif ($opt eq 'delete')
	{

		# loop through phones, delete from db indexed by phone number
		while (@{$phones})
		{
			my $index  = shift(@{$phones});
			my $status = $db->db_del($index);

			print LOGGER localtime() . ":: del status: $status\n";
		}

	}
	elsif ($opt eq 'clear')
	{

		# add 1 element to db, just so there is always something to truncate
		# returns number of elements truncated. failure returns 0
		my $count;
		$db->db_put(1, 1);
		$db->truncate($count);

		print LOGGER localtime() . ":: clear status: $count\n";

		# if error
		unless ($count)
		{
			my $result->{'faultString'} = "Could not clear db";

			$db->db_close();
			close(LOGGER);

			return $result;
		}
	}
	elsif ($opt eq 'get')
	{

		my $result;

		# loop through phones, grab data, index by phone number
		# ext and descriptions are data separated by the first colon
		# a colon is appropriate because
		# there will never be a colon in the ext number to muck things up

		while (@{$phones})
		{
			my $index = shift(@{$phones});

			my $value;
			my $status = $db->db_get($index, $value);

			my ($ext_result, $desc_result) = split(/:/, $value, 2);
			$result->{$index}->{'ext'}  = $ext_result;
			$result->{$index}->{'desc'} = $desc_result;

			print LOGGER localtime()
			  . ":: get ext: $ext_result, get desc: $desc_result\n";

			# if error
			if ($status)
			{
				$result->{'faultString'} = "$index does not exist in db";

				$db->db_close();
				close(LOGGER);

				return $result;
			}
		}

		$db->db_close();
		close(LOGGER);

		return $result;
	}

	# invalid opt
	else
	{

		#close file
		$db->db_close();
		close(LOGGER);
		return 0;
	}

	$db->db_close();
	close(LOGGER);
	return 1;

}

=head2 stringify_args

Turn function arguments into strings with a particular focus
on the case where the arguments come as an ArrayRef.

       Args: a function name and its arguments
    Returns: a string of arg name and values for printing to xpl.log
    Example: stringify_args()

=cut

sub stringify_args
{

	# shift off function name
	# which means @_ now should be original args to the function
	my $function_name = shift;

	# return @_ joined if the first element, $_[0], is not any ArrayRef
	# LIMITATION: This technique only processes the first ArrayRef, i.e $_[0].
	# However, that seems to be the pattern of usage in passing args to funcs.cgi subroutine
	return join ', ', @_ if !ref($_[0]) || (ref($_[0]) ne 'ARRAY');
	my $args_ref = $_[0];

	# NOTE: Adjust length of values returned.
	my $MAX_VALUE_LENGTH = 32;

	# If we have an ArrayRef, then build a log string
	# using the &param_map to couple param name with value
	# $_[0]->[$i] where $i = 0, 1, 2 ... represent the value
	# and the name comes from &param_map->{$function_name}->{$i}
	return join ';  ', map {
		&param_map->{$function_name}->{$_} . ": " . substr $args_ref->[$_], 0,
		  $MAX_VALUE_LENGTH
	  }
	  sort keys %{ &param_map->{$function_name} };
}

=head2 param_map

Define the argument names and values that are passed
to certain subroutines within funcs.cgi.  This map
is used to improve the info written to xpl.log.

    Arguments: None
      Returns: HashRef[HashRef[Int]], i.e. the structure below
      Example: my $param_map = &param_map;
	           my $url = $param_map->{copy_func}->{0};

=cut

sub param_map
{

	return {
		copy_func  => { 0 => 'url', 1 => 'path', 2 => 'md5', },
		forward_vm => {
			0 => 'email',
			1 => 'file',
			2 => 'fromname',
			3 => 'fromemail',
			4 => 'subject',
			5 => 'body',
			6 => 'unbound_sid',
		},
		get_reports  => { 0 => 'server_id', 1 => 'uae' },
		ls           => { 0 => 'path',      1 => 'directory' },
		mv           => { 0 => 'from',      1 => 'to' },
		restart      => { 0 => 'restart_type' },
		server_stats => { 0 => 'stat_time', 1 => 'stat_type' },
		sync_config  => {
			0 => 'file',
			1 => 'md5',
			2 => 'data',
			3 => 'path',
			4 => 'noreload',
			5 => 'mosted_ok'
		},
	};
}

=head2 log_args

Write function paramters to xpl.log.

          Args: function argument
       Returns:	void
   Side Effect: Write function argument names and values to xpl.log
       Example: sub ls {
	    log_args(@_);
	    # ...
	  }
    Limitations: Uses caller(1)[3] which is the function
	from which the log_args() was called, i.e. parent.

=cut

sub log_args
{
	# Get name of sub which called this log_args()
	# Expect the form of 'main::$fucntion_name, e.g.
	# main::ls
	my $parent_sub = (caller(1))[3];
	my ($function_name) = $parent_sub =~ m/\:?(\w+)$/;

	open(LOG, ">>/tmp/xpl.log");
	print LOG localtime() 
	  . ": $$: "
	  . $function_name . '.xpl('
	  . stringify_args($function_name, @_) . ")\n";
	close(LOG);
}



##############################################################################
# recurse: Reads all files within a directory recursively
#
#    Args: file path
# Returns: list of files
##############################################################################
sub recurse
{
	#open(LOG, ">>/tmp/xpl.log"); print LOG localtime() . ": $$: recurse.pl(@_)\n"; close(LOG);
    my $path = shift;

    my @files;

    opendir(PATH, $path);
    foreach (grep { $_ ne "." && $_ ne ".." } readdir(PATH))
    {
        if (-d "$path/$_")
        {
            push @files, recurse("$path/$_");
        }

        push @files, "$path/$_";
    }

    return @files;
}


##############################################################################
# read_config: Reads include files and all other includes within
#
#    Args: file path
# Returns: reference to array of lines of the config
##############################################################################
sub read_config
{
	#open(LOG, ">>/tmp/xpl.log"); print LOG localtime() . ": $$: read_config.pl(@_)\n"; close(LOG);
    my $file = shift;
    my $files = shift || {};
    $files->{$file} = 1;

    my @data;

    open(FILE, "<$file") || die "Can't read $file: $!";
    my @file = <FILE>;
    foreach (@file)
    {
        if (/^#include\s*"(.*)"/)
        {
            # prevent recursive loops
            next if $files->{$1};
            push @data, @{read_config($1, $files)};
        }
        else
        {
            push @data, $_;
        }
    }
    close FILE;

    return \@data;
}


##############################################################################
# strip_nasties: Strip any bad characters out of a string
#
#    Args: text
# Returns: clean_text
##############################################################################
sub strip_nasties {
	#open(LOG, ">>/tmp/xpl.log"); print LOG localtime() . ": $$: strip.pl(@_)\n"; close(LOG);
    my($txt) = shift(@_);
    my($OK_CHARS) ='-a-zA-Z0-9_.@';
    $txt =~ s/[^$OK_CHARS]/-/go;
    return($txt);
}

