<?xml version="1.0"?>
<!DOCTYPE methoddef SYSTEM "rpc-method.dtd">
<!--
    Generated automatically by make_method v1.09, Sat Sep  4 08:46:54 2004

    Any changes made here will be lost.
-->
<methoddef>
<name>uae_routing</name>
<version>1.0</version>
<signature>array array array array array</signature>
<help>
Returns success/failure or data hash, based on the operation
</help>
<code language="perl">
<![CDATA[
#!/opt/perl5/bin/perl
###############################################################################
#
#   Sub Name:       uae_routing
#
#   Description:    Manage Call Routing Data
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#
#   Globals:        None.
#
#   Environment:    None.
#
#   Returns:        Success/failure/data hash
#
###############################################################################
sub uae_routing
{
	my ($srv, $req, $phones, $exts, $descs) = @_;
	my ($sid) = shift(@{$req});
	my ($opt) = shift(@{$req});

	use strict;
	use File::Copy;
	use BerkeleyDB;

	my $db = new BerkeleyDB::Hash
		(
			-Filename => "/var/lib/asterisk/call_routing.db",
			-Flags    => DB_CREATE
		)
		or die "Could not open DB";

	open(LOGGER, ">>/tmp/test_routing.log") or die "Can't open file: $!";

	# adding
	if($opt eq 'add' || $opt eq 'modify')
	{
		# loop through phones, put into db indexed by phone number
		# (assumes array of phones, exts, desc are of equal length)
		while(@{$phones})
		{	

			my $index = shift(@{$phones});
			my $record = shift(@{$exts}) . ":" . shift(@{$descs});

			my $status = $db->db_put($index, $record);

			print LOGGER localtime() . ":: add status: $status, adding: $index : $record\n";

			# if error
			if($status)
			{
				my $result;
				$result->{'faultString'} = "Could not add ($index, $record)";	

				$db->db_close();
				close(LOGGER);
				
				return $result;
			}
		}
	}	
	elsif($opt eq 'delete')
	{
		# loop through phones, delete from db indexed by phone number
		while(@{$phones})
		{
			my $index = shift(@{$phones});
			my $status = $db->db_del($index);

			print LOGGER localtime() . ":: del status: $status\n";
		}

	}
	elsif($opt eq 'clear')
	{
		# add 1 element to db, just so there is always something to truncate
		# returns number of elements truncated. failure returns 0
		my $count;
		$db->db_put(1,1);
		$db->truncate($count);

		print LOGGER localtime() . ":: clear status: $count\n";

		# if error
		unless($count)
		{
			my $result;
			$result->{'faultString'} = "Could not clear db";

			$db->db_close();
			close(LOGGER);

			return $result;
		}
	}
	elsif($opt eq 'get')
	{

		my $result;

		# loop through phones, grab data, index by phone number
		# ext and descriptions are data separated by the first colon
		# a colon is appropriate because
		# there will never be a colon in the ext number to muck things up

		while(@{$phones})
		{
			my $index = shift(@{$phones});

			my $value;
			my $status = $db->db_get($index, $value);
		

			my ($ext_result, $desc_result) = split(/:/, $value, 2);
			$result->{$index}->{'ext'} = $ext_result;
			$result->{$index}->{'desc'} = $desc_result;

			print LOGGER localtime() . ":: get ext: $ext_result, get desc: $desc_result\n";

			# if error
			if($status)
			{
				$result->{'faultString'} = "$index does not exist in db";

				$db->db_close();
				close(LOGGER);

				return $result;
			}
		}

		$db->db_close();
		close(LOGGER);
	
		return $result;
	}
	# invalid opt
	else
	{
		#close file
		$db->db_close();
		close(LOGGER);
		return 0;
	}

	$db->db_close();
	close(LOGGER);
	return 1;

}

__END__
]]></code>
</methoddef>



